@@ -1,5 +1,6 @@
 import os
-from typing import Generator
+import json
+from typing import Any, Generator
 
 
 def read_resource_file(package_path: str, resource_name: str, strip: bool = True) -> Generator[str, None, None]:
@@ -20,3 +21,17 @@ def read_resource_file(package_path: str, resource_name: str, strip: bool = True
             lines = (line.strip() for line in lines)
         lines = (line for line in lines if line and not line.startswith('#'))
         return lines
+
+
+def read_resource_json_file(package_path: str, resource_name: str) -> Any:
+    """
+    Load a resource file using json
+
+    :param package_path: package path where the resource is located
+    :param resource_name: path to the resource, relative from package
+    :return: Contents of the jsoned file
+    """
+    # Read the document and remove comment lines
+    resource_path = os.path.join(os.path.dirname(package_path), resource_name)
+    with open(resource_path, 'r') as f:
+        return json.load(f)
